//_ minwinmain.d   Tue Nov 05 2002 */
// Copyright (c) 2002-2004 by Walter Bright
// All Rights Reserved
// Compile with Digital Mars compiler www.digitalmars.com
// www.classicempire.com

// Ported from winmain.d to use MinWin by Ben Hinkle

import std.c.stdlib;
import std.string;

private {
  import minwin.all;
  import minwin.logging;
  alias minwin.text.Text MWText;
  import empire;
  import winemp;
  import eplayer;
  import display;
  import twin;
  import init;
  import var;
  import maps;
  import move;
  import text;
}
alias text.Text Text;

version(Windows) {
  pragma(lib,"minwin");
  private import std.c.windows.windows; // for PlaySound
  alias minwin.geometry.RGB RGB; // resolve conflict

} else version(Motif) {

  // these don't work yet
  pragma(lib,"minwin_motif");
  pragma(lib,"Xm");
  pragma(lib,"Xt");
  pragma(lib,"X11");
  pragma(lib,"Imlib");

  // seems like build program uses a different pragma
  //  version(build) pragma(link,Xm);
  //  version(build) pragma(link,Xt);
  //  version(build) pragma(link,X11);
  //  version(build) pragma(link,Imlib);

} else version(GTK) {
  pragma(lib,"minwin_gtk");
  pragma(lib,"gtk-x11-2.0");
  pragma(lib,"X11");
}

/* Collect all Windows static global data.
 */

struct Global
{
  EmpireWindow win;
  int inited;		// !=0 means game is initialized
  Image hSplash;

  // City Select
  int phase;
  int newphase;
  SelectCityDialog selectCityDialog;

  // Init
  int numplayers;
  int newnumplayers;
  int demo;

  int speaker;	// !=0 means sound is on

  double scalex;	// zoom factor
  double scaley;	// zoom factor

  // Font
  Font hFont;
  short cxChar, cxCaps, cyChar, cyAscent;
  int cyOffset;

  // Pen
  Pen hPen;

  // Bitmaps
  Image mapvaltab[MAPMAX];
  Image unknown10;

  Player *player;	// which player is being displayed
  ubyte *map;		// which map is being displayed
  loc_t ulcorner;	// upper left corner
  loc_t cursor;	// location of cursor
  Image hCursor;
  int offsetx;
  int offsety;

  // Window size
  int cxClient, cyClient;

  // Clipping rectangles
  Rect sector;
  Rect text;

  Region sectorRegion;
  Region textRegion;

  // Sector size
  int pixelx;
  int pixely;

  // Blast
  Image hBlast;
  Image hBlastmask;
  int blastState;	// !=0 means draw blast
  int blastx;
  int blasty;		// location of blast
};

Global global;


extern(C)
int MinWinMain(Application* app)
{
    static   char szAppName [] = "Empire";

    // TODO: icons
    //        wndclass.hIcon         = LoadIconA(hInstance, "About");

    Window win = new EmpireWindow("Empire: Wargame of the Century",szAppName);
    win.size = XY(500,500);
    win.visible = true;

    app.idleDelegate ~= delegate void() {
      if (global.inited) {
	slice();
      }
    };
    app.idleTime(10);
    return app.enterEventLoop();
}

void DrawBitmap(GContext gc, short xStart, short yStart, 
		Image hBitmap, double scalex, double scaley, 
		PaintMode mode)
{
  gc.stretchImage(hBitmap,xStart,yStart,
		  cast(int) (hBitmap.width * scalex + .99),
		  cast(int) (hBitmap.height * scaley + .99),
		  mode);
}

class EmpireWindow : Window {
  NewGameDialog newGameDialog;

  this(char[] title, char[] name) {
    super(title,name);

    // initial size?
    
    global.speaker = 1;

    global.cxClient = 120;
    global.cyClient = 160;
    this.size = XY(120,160);

    global.pixelx = 120;
    global.pixely = 120;

    global.win = this;
    global.scalex = 1.0;
    global.scaley = 1.0;
    global.numplayers = IDD_FOUR;
    global.map = .map;
    global.offsetx = 0;
    global.offsety = 0;

    /*	    
    FontData fd;
    fd.size = 10;
    fd.name = "Courier";
    global.hFont = new Font(fd);
    */
    global.hFont = standardFont(StandardFont.Fixed);
    GContext gc = getGContext();
    Font oldfont = gc.setFont(global.hFont);
    FontMetrics fm;
    gc.getFontMetrics(fm);

    global.cxChar = fm.maxWidth;
    global.cxCaps = global.cxChar;
    global.cyChar = fm.size+fm.leading;
    global.cyAscent = fm.ascent;
    gc.dispose();

    makeMenus();

    int x;
    getPeerOffset(x,global.cyOffset);

    // Clipping rectangles
    global.text.LTRB(0,0,global.pixelx,4*global.cyChar+global.cyOffset);

    global.sector.LTRB(0,4*global.cyChar+global.cyOffset,global.pixelx,
		       global.sector.top + global.pixely);

    global.sectorRegion = new Region(global.sector);
    global.textRegion = new Region(global.text);

    version(Windows) {
      // use the existing mapping defined in empire.rc
      for (int i = 0; i < MAPMAX; i++)
    	{
    	  global.mapvaltab[i] = loadCompatibleImage("map" ~ .toString(i+1));
    	}
    } else {
      global.mapvaltab[0] = loadCompatibleImage("unknown");
      global.mapvaltab[1] = loadCompatibleImage("city");
      global.mapvaltab[2] = loadCompatibleImage("sea");
      global.mapvaltab[3] = loadCompatibleImage("land");
      for (int i = 0; i < 6; i++) {
	global.mapvaltab[4+10*i] = loadCompatibleImage("city"~.toString(i+1));
	version(LOG)log.printf("loaded %d %p\n",i,global.mapvaltab[4+10*i]);
	global.mapvaltab[5+10*i] = loadCompatibleImage("a"~.toString(i+1));
	global.mapvaltab[6+10*i] = loadCompatibleImage("f"~.toString(i+1));
	global.mapvaltab[7+10*i] = loadCompatibleImage("fs"~.toString(i+1));
	global.mapvaltab[8+10*i] = loadCompatibleImage("d"~.toString(i+1));
	global.mapvaltab[9+10*i] = loadCompatibleImage("t"~.toString(i+1));
	global.mapvaltab[10+10*i] = loadCompatibleImage("s"~.toString(i+1));
	global.mapvaltab[11+10*i] = loadCompatibleImage("r"~.toString(i+1));
	global.mapvaltab[12+10*i] = loadCompatibleImage("c"~.toString(i+1));
	global.mapvaltab[13+10*i] = loadCompatibleImage("b"~.toString(i+1));
      }
    }
    global.unknown10 = loadCompatibleImage("unknown10");
    global.hCursor = loadCompatibleImage("cursor");
    global.hSplash = loadCompatibleImage("splash");
    global.hBlast = loadCompatibleImage("blast");
    global.hBlastmask = loadCompatibleImage("blastmask");

    keyDelegate ~= &doKey;
    paintDelegate ~= &doPaint;
    windowDelegate ~= &doWindowEvent;
    quitOnDestroy = true;
  }

  void makeMenus() {
    MenuBar mb = new MenuBar(this);
    Menu file;
    with (file = new Menu(mb,"File")) {
      add("New Game...",IDM_NEW);
      add("Open Game...",IDM_OPEN);
      add("Save Game...",IDM_SAVE);
      addSeparator();
      add("Exit",IDM_CLOSE);
    }

    Menu view;
    with (view = new Menu(mb,"View")) {
      add("+ Zoom in",IDM_ZOOMIN);
      add("- Zoom out",IDM_ZOOMOUT);
      add("N Center screen",IDM_N);
      add("O Change POV",IDM_POV);
    }

    Menu cmd;
    with (cmd = new Menu(mb,"Cmd")) {
      add( "F From To",		IDM_F);
      add( "G Goto city",		IDM_G);
      add( "H 20 free moves",	IDM_H);
      add( "I Direction",		IDM_I);
      add( "J Sound on/off",	IDM_SOUND);
      add( "K Wake up",		IDM_K);
      add( "L Load A/F",		IDM_L);
      add( "P City production",	IDM_P);
      add( "R Move randomly",	IDM_R);
      add( "S Sentry",		IDM_S);
      add( "U Wake up A/F",	IDM_U);
      add( "Y Survey",		IDM_Y);
      add( "Esc Exit mode",	IDM_ESC);
      add( "< Faster",		IDM_FASTER);
      add( "> Slower",		IDM_SLOWER);
    }

    Menu help;
    with (help = new Menu(mb,"Help")) {
      add("Help",IDM_HELP);
      add("About Empire...",IDM_ABOUT);
    }
  }

  void doKey(Component source, KeyEvent* event) {
    if (event.id == KeyPressedEvent) {
      dchar ch = event.keyChar;
      switch (ch) {
      case 'j':
      case 'J':
	global.speaker ^= 1;
	return 0;
	
      case 12:
	repaint();
	break;
	
      default:
	insert(ch);
	break;
      }
    }
    // also handle WM_KEYDOWN?
  }

  FileFilter[] szFilter = [{"Empire Files", ["emp"]} ];

  void doCommand(int cmd) { // for menu items 
    int ch;
    switch(cmd) {
    case IDM_NEW:		// start new game
      if (newGameDialog is null) 
	newGameDialog = new NewGameDialog(this);
      newGameDialog.setData(global.demo,global.numplayers);
      newGameDialog.visible = true;
      if (newGameDialog.ok == true) {
	newGameDialog.getData(global.demo,global.numplayers);
	init_var();
	winSetup();
	global.inited = 1;
      }
      repaint();
      return 0;

    case IDM_OPEN:		// open saved game
      FileDialogData data;
      data.title = "Open game file";
      data.filter = szFilter;
      if (openFileDialog(this,data)) {
	FILE *fp;

	fp = fopen(data.result, "rb");
	if (!fp) {
	  errorDialog(this,"Could not read EMP file","Empire Restore");
	} else {
	    init_var();
	    if (resgam(fp))
	      {
	      errorDialog(this,"Corrupt EMP file","Empire Restore");
	      winSetup();
	      }
	    else
	      winRestore();
	    global.inited = 1;
	    global.win.repaint();
	}
      }
      return 0;

    case IDM_SAVE:		// save existing game
      FileDialogData data;
      data.title = "Save game file";
      data.filter = szFilter;
      if (saveFileDialog(this,data)) {
	if (var_savgam(data.result)) {
	  errorDialog(this,"Could not write EMF file","Empire Save");
	}
      }
      return 0;

    case IDM_CLOSE:
      gApp.exitEventLoop();
      return 0;

    case IDM_SOUND:
      global.speaker ^= 1;
      return 0;

    case IDM_ABOUT:
      showAboutBox();
      return 0;

    case IDM_HELP:
      help();
      return 0;

    case IDM_ZOOMIN:
      zoomin();

    case IDM_ZOOMOUT:
      zoomout();

    case IDM_F:    ch = 'F';	insert(ch); break;
    case IDM_G:    ch = 'G';	insert(ch); break;
    case IDM_H:    ch = 'H';	insert(ch); break;
    case IDM_I:    ch = 'I';	insert(ch); break;
    case IDM_K:    ch = 'K';	insert(ch); break;
    case IDM_L:    ch = 'L';	insert(ch); break;
    case IDM_N:    ch = 'N';	insert(ch); break;
    case IDM_P:    ch = 'P';	insert(ch); break;
    case IDM_R:    ch = 'R';	insert(ch); break;
    case IDM_S:    ch = 'S';	insert(ch); break;
    case IDM_U:    ch = 'U';	insert(ch); break;
    case IDM_Y:    ch = 'Y';	insert(ch); break;
    case IDM_ESC:  ch = ESC;	insert(ch); break;
    case IDM_FASTER:    ch = '<';	insert(ch); break;
    case IDM_SLOWER:    ch = '>';	insert(ch); break;
    case IDM_POV:    ch = 'O';	insert(ch); break;

    default:
      break;
    }
  }
  void insert(int ch) {
    //    version(LOG)log.writefln("got char %x",ch);
    // Insert into buffer of player we are watching
    Player *p;
    
    for (int i = 1; i <= numply; i++) {
      p = Player.get(i);
      if (p.watch) {
	p.display.text.TTunget(ch);
	break;
      }
    }
  }

  void zoomin() {
    // TODO
  }
  void zoomout() {
    // TODO
  }

  void doWindowEvent(Component source, WindowEvent* event) {
    //    version(LOG)log.writefln("WindowEvent %x",event.id);
    if (event.id == WindowSizeEvent  ||event.id == WindowSizingEvent ) {
      Rect r = this.bounds;
      global.cyClient = r.height;
      global.cxClient = r.width;

      global.pixelx = global.cxClient;
      if (global.pixelx < 120)
	global.pixelx = 120;
      if (global.pixelx > (Mcolmx + 1) * 10)
	global.pixelx = (Mcolmx + 1) * 10;
      if (global.pixelx / cast(int)(10 * global.scalex) < 5)
	{
	  global.scalex = global.pixelx / (10 * 5.0);
	  global.scaley = global.scalex;
	}
      global.pixely = global.cyClient - 4*global.cyChar;
      if (global.pixely < 120)
	global.pixely = 120;
      if (global.pixely > (Mrowmx + 1) * 10)
	global.pixely = (Mrowmx + 1) * 10;
      if (global.pixely / cast(int)(10 * global.scaley) < 5)
	{
	  global.scaley = global.pixely / (10 * 5.0);
	  global.scalex = global.scaley;
	}

      global.text.right = global.pixelx;
      global.sector.LTRB( global.sector.left, global.sector.top, 
			  global.pixelx,
			  global.sector.top + global.pixely);
      global.sectorRegion = new Region(global.sector);
      //      r.LTRB( global.text.left, global.text.top, global.text.right, global.text.bottom);
      global.textRegion = new Region(global.text);

      if (global.inited && global.player)
	{	Display *d = global.player.display;

	d.secbas = -1;
	d.setdispsize(d.text.nrows, d.text.ncols);
	d.text.clear();
	adjSector(global.scalex, global.scaley);
	}
    }
  }

  void doPaint(Component source, GContext gc) {

    //    version(LOG)log.printf("in paint %d %x\n",cast(int)global.inited,cast(int)global.player);
    if (!global.inited || !global.player)
      {	double sx, sy;
      /*
	sx = global.cxClient / 120.0;
	if (sx < 1)
	sx = 1;
	sy = global.cyClient / 160.0;
	if (sy < 1)
	sy = 1;
      */
      Point sz = this.size;
      gc.stretchImage(global.hSplash,0,0,sz.x,sz.y, PaintMode.Copy);
      version (Windows) {
	static int intro;
	if (!intro++)
	  PlaySoundA("intro.wav", null, SND_ASYNC | SND_FILENAME);
      }
      }
    else
      {
	//PRINTF("WM_PAINT: ulcorner = %d, cursor = %d, offsetx = %d, offsety = %d\n", global.ulcorner, global.cursor, global.offsetx, global.offsety);
	int r, c;
	int dx;
	int dy;
	PaintMode mode;
	Rect clipbox = gc.updateRect;
	//PRINTF("sector : %2d,%2d %2d,%2d\n", global.sector.left, global.sector.top, global.sector.right, global.sector.bottom);
	//PRINTF("clipbox: %2d,%2d %2d,%2d\n", clipbox.left, clipbox.top, clipbox.right, clipbox.bottom);
	if (clipbox.bottom < global.sector.top)
	  goto LpaintText;

	//TODO		gc.setClip(global.sectorRegion);
	//SelectClipRgn(hdc, global.sectorRegion);

	r = ROW(global.ulcorner);
	c = COL(global.ulcorner);
	int rmax, cmax;
	dx = cast(int)(10 * global.scalex);
	dy = cast(int)(10 * global.scaley);
	rmax = r + (global.offsety + global.pixely + dy - 1) / dy;
	cmax = c + (global.offsetx + global.pixelx + dx - 1) / dx;
	if (rmax > Mrowmx)
	  rmax = Mrowmx + 1;
	if (cmax > Mcolmx)
	  cmax = Mcolmx + 1;
	//		version(LOG)log.printf("in paint not in splash mode %d %d %d %d\n",r,c,rmax,cmax);
	for (int j = r; j < rmax; j++)
	  {   int y;

	  y = global.sector.top + (j - r) * dy - global.offsety;
	  if (y >= clipbox.bottom ||
	      y + dy < clipbox.top)
	    continue;

	  for (int i = c; i < cmax; i++)
	    {
	      loc_t loc = j * (Mcolmx + 1) + i;
	      Image h;
	      int x;

	      x =  (i - c) * dx - global.offsetx;
	      if (x >= clipbox.right ||
		  x + dx < clipbox.left)
		continue;

	      h = global.mapvaltab[global.map[loc]];
	      if ((j % 10) == 0 && (i % 10) == 0 &&
		  global.map[loc] == 0)
		h = global.unknown10;
	      mode = PaintMode.Copy;
	      if (loc == global.cursor && global.player.mode != mdSURV)
		{
		  mode = PaintMode.Invert;
		}
	      //			version(LOG)log.writefln("DrawBitmap %d %d %s %s %s",x,y,h,
	      //						 global.scalex,global.scaley);
	      DrawBitmap(gc, x, y, h,
			 global.scalex, global.scaley, mode);
	    }
	  }

	// Draw a rectangle around the map edge
	int x1,y1,x2,y2;
	x1 = -c * dx - global.offsetx;
	y1 = 40 - r * dx - global.offsety;
	x2 = x1 + (Mcolmx + 1) * dx - 1;
	y2 = y1 + (Mrowmx + 1) * dy - 1;
	PenData pd;
	pd.width = dx/3+2;
	pd.color = RGB(255,0,0);
	global.hPen = new Pen(&pd);
	Pen oldPen = gc.setPen(global.hPen);
	Point[4] pts;
	pts[0] = XY(x1,y1);
	pts[1] = XY(x2,y1);
	pts[2] = XY(x2,y2);
	pts[3] = XY(x1,y2);
	gc.drawPolygon(pts);
	gc.setPen(oldPen);
	delete global.hPen;

	// Do the blast graphic
	if (global.blastState)
	  {
	    DrawBitmap(gc, global.blastx, global.blasty, 
		       global.hBlastmask, 1.0, 1.0, 
		       PaintMode.And);
	    DrawBitmap(gc, global.blastx, global.blasty, 
		       global.hBlast, 1.0, 1.0, 
		       PaintMode.Or);
	  }

	// Do the survey mode graphic
	if (global.player && global.player.mode == mdSURV)
	  {
	    int x1,y1,x2,y2;
	    int cursorx, cursory;

	    PenData pd;
	    pd.style = PenStyle.Dash;
	    pd.color = RGB(255,255,255);
	    auto Pen hPen = new Pen(&pd);
	    Pen oldPen = gc.setPen(hPen);

	    cursorx = LocToX(global.player.curloc);
	    cursory = LocToY(global.player.curloc);

	    x1 = 0;
	    y1 = cursory;
	    x2 = cursorx - dx/2;
	    y2 = y1;
	    gc.drawLine(x1,y1,x2,y2);

	    x1 = x2 + dx;
	    x2 = global.pixelx;
	    gc.drawLine(x1,y1,x2,y2);

	    x1 = cursorx;
	    y1 = 40;
	    x2 = x1;
	    y2 = cursory - dy/2;
	    gc.drawLine(x1,y1,x2,y2);

	    y1 = y2 + dy;
	    y2 = 40 + global.pixely;
	    gc.drawLine(x1,y1,x2,y2);

	    gc.setPen(oldPen);
	  }

	// Do the direction mode graphic
	if (global.player && global.player.mode == mdDIR)
	  {
	    int x1,y1,x2,y2;
	    int cursorx, cursory;

	    PenData pd;
	    pd.style = PenStyle.Dash;
	    pd.color = RGB(255,255,255);
	    auto Pen hPen = new Pen(&pd);
	    Pen oldPen = gc.setPen(hPen);

	    cursorx = LocToX(global.player.curloc);
	    cursory = LocToY(global.player.curloc);

	    /* -O */
	    x1 = cursorx - dx/2 - 2 * dx;
	    y1 = cursory;
	    x2 = cursorx - dx/2;
	    y2 = y1;
	    gc.drawLine(x1,y1,x2,y2);

	    /* \
	     *  O
	     */
	    y1 -= 2 * dy + dy/2;
	    y2 -= dy/2;
	    gc.drawLine(x1,y1,x2,y2);

	    /*  O
	     * /
	     */
	    y1 += dy * 5;
	    y2 += dy;
	    gc.drawLine(x1,y1,x2,y2);

	    /*  O
	     *  |
	     */
	    x1 = cursorx;
	    x2 = x1;
	    gc.drawLine(x1,y1,x2,y2);

	    /*  O
	     *   \
	     */
	    x1 += (dx - dx/2) + 2 * dx;
	    x2 += (dx - dx/2);
	    gc.drawLine(x1,y1,x2,y2);

	    /*  O- */
	    y1 = cursory;
	    y2 = y1;
	    gc.drawLine(x1,y1,x2,y2);

	    /*   /
	     *  O
	     */
	    y1 -= dy/2 + 2 * dy;
	    y2 -= dy/2;
	    gc.drawLine(x1,y1,x2,y2);

	    /*  |
	     *  O
	     */
	    x1 = cursorx;
	    x2 = x1;
	    gc.drawLine(x1,y1,x2,y2);

	    gc.setPen(oldPen);
	  }

	// Do the move mode graphic
	if (global.player && global.player.mode == mdTO)
	  {  
	    int x1,y1,x2,y2;

	    PenData pd;
	    pd.style = PenStyle.Dash;
	    pd.color = RGB(255,255,255);
	    auto Pen hPen = new Pen(&pd);
	    Pen oldPen = gc.setPen(hPen);

	    x1 = LocToX(global.player.frmloc);
	    y1 = LocToY(global.player.frmloc);
	    x2 = LocToX(global.player.curloc);
	    y2 = LocToY(global.player.curloc);

	    if (x1 != x2 && y1 != y2)
	      {	int x, y;
	      int ax, ay;

	      ax = abs(x1 - x2);
	      ay = abs(y1 - y2);
	      if (ax < ay)
		{
		  x = x2;
		  if (y1 < y2)
		    y = y1 + ax;
		  else
		    y = y1 - ax;
		}
	      else
		{
		  if (x1 < x2)
		    x = x1 + ay;
		  else
		    x = x1 - ay;
		  y = y2;
		}
	      gc.drawLine(x1,y1,x,y);
	      gc.drawLine(x,y,x2,y2);
	      } else {
		gc.drawLine(x1,y1,x2,y2);
	      }

	    gc.setPen(oldPen);
	  }

      LpaintText:
	if (clipbox.bottom > global.text.top &&
	    clipbox.top < global.text.bottom)
	  {
	    // Do the text box
	    //TODO		  gc.setClip(global.textRegion);
	    //		    SelectClipRgn(hdc, global.textRegion);

	    // Fill background
	    auto Brush b = new Brush(RGB(255,255,255));
	    Brush oldBrush = gc.setBrush(b);
	    gc.fillRect(global.text);
	    gc.setBrush(oldBrush);

	    pd.color = RGB(0,0,0);
	    auto Pen hPen = new Pen(&pd);
	    oldPen = gc.setPen(hPen);
	    Font oldFont = gc.setFont(global.hFont);
	    for (int i = 0; i < 4; i++)
	      {
		char* cstr = vbuffer[i];
		char[] str = cstr[0 .. strlen(cstr)];
		int offset = global.cyOffset;
		// Motif draws text wrt baseline! todo: move to minwin
		version(Motif)
		  offset += global.cyAscent; 
		gc.drawText(0,global.cyChar*i+offset, str);
	      }
	    gc.setFont(oldFont);
	    gc.setPen(oldPen);
	  }
      }
  }
}

/********************************************
 * "City Select" dialog box.
 */


/********************************
 * Dialog box to get city phase.
 * Returns:
 *	new phase
 */

int dialogCitySelect(int oldphase)
{
  global.win.repaintNow();
  if (oldphase & ~7)
    oldphase = 0;		// default to armies
  global.phase = oldphase + IDD_ARMIES;
  if (global.selectCityDialog is null) 
    global.selectCityDialog = new SelectCityDialog(global.win);
  global.selectCityDialog.setData(global.phase);
  global.selectCityDialog.visible = true;
  global.selectCityDialog.getData(global.phase);
  return global.phase - IDD_ARMIES;
}

/******************************
 * Flush the display.
 */

extern (C) void win_flush()
{
  global.win.repaint(global.text);
  global.win.repaintNow();
}

/******************************
 * Click the speaker.
 */

extern (C) void sound_click()
{
  global.win.repaintNow();
  version (Windows) {
    if (global.speaker)
  	PlaySoundA("click.wav", null, SND_ASYNC | SND_FILENAME | SND_NOSTOP);
  }
}

void sound_gun()
{
  global.win.repaintNow();
  version (Windows) {
    if (global.speaker)
      PlaySoundA("gun_1.wav", null, SND_SYNC | SND_FILENAME);
  }
}

void sound_bang()
{
  global.win.repaintNow();
  version (Windows) {
      if (global.speaker)
      {	PlaySoundA("explosi1.wav", null, SND_SYNC | SND_FILENAME);
  	PlaySoundA("bubbles.wav", null, SND_SYNC | SND_FILENAME);
      }
  }
}

void sound_error()
{
  global.win.repaintNow();
  version (Windows) {
      if (global.speaker)
  	PlaySoundA("error.wav", null, SND_SYNC | SND_FILENAME);
  }
}

void sound_splash()
{
  global.win.repaintNow();
  version (Windows) {
      if (global.speaker)
  	PlaySoundA("splash.wav", null, SND_SYNC | SND_FILENAME);
  }
}

void sound_aground()
{
  global.win.repaintNow();
  version (Windows) {
      if (global.speaker)
  	PlaySoundA("bubbles.wav", null, SND_SYNC | SND_FILENAME);
  }
}

void sound_subjugate()
{
  global.win.repaintNow();
  version (Windows) {
      if (global.speaker)
  	PlaySoundA("machine1.wav", null, SND_SYNC | SND_FILENAME);
  }
}

void sound_crushed()
{
  global.win.repaintNow();
  version (Windows) {
      if (global.speaker)
  	PlaySoundA("gun_3.wav", null, SND_SYNC | SND_FILENAME);
  }
}

void sound_flyby()
{
  global.win.repaintNow();
  version (Windows) {
      if (global.speaker)
  	PlaySoundA("flyby.wav", null, SND_SYNC | SND_FILENAME);
  }
}

void sound_fcrash()
{
  global.win.repaintNow();
  version (Windows) {
      if (global.speaker)
  	PlaySoundA("explode.wav", null, SND_SYNC | SND_FILENAME);
  }
}

void sound_fuel()
{
  global.win.repaintNow();
  version (Windows) {
      if (global.speaker)
  	PlaySoundA("fuel.wav", null, SND_SYNC | SND_FILENAME);
  }
}

void sound_taps()
{
  global.win.repaintNow();
  version (Windows) {
      if (global.speaker)
  	PlaySoundA("taps.wav", null, SND_SYNC | SND_FILENAME);
  }
}

void sound_ackack()
{
  global.win.repaintNow();
  version (Windows) {
      if (global.speaker)
  	PlaySoundA("ackack1.wav", null, SND_SYNC | SND_FILENAME);
  }
}

/***********************************
 * Setup for Windows.
 * Equivalent to main.setup().
 */

void winSetup()
{
    //PRINTF("winSetup()\n");
    debug
    {
	// set random number generator to predictable value
	setran();
    }

    Text *t = &player[0].display.text;

    //printf("Please wait seven days for creation of world...\n");
    selmap();			// read in map
    citini();			// init city variables

    numply = global.numplayers - IDD_ONE + 1;
    numleft = numply;
    for (plynum = 0; plynum <= numply; plynum++)
    {
//PRINTF("player %d\n", plynum);
	Player *p = &player[plynum];
	p.display = new Display();
	Display *d = p.display;
	d.initialize();

	p.num = plynum;
	p.map = (plynum == 0) ? .map : cast(ubyte *)calloc(MAPSIZE,1);
	p.human = (plynum == 1 && !global.demo);
	p.watch = DAnone;

	if (p.human)
	{
	    d.timeinterval = 1;
	}
	else
	{
	}

	if (plynum == 1)
	{
	    p.secflg = 1;
	    p.watch = DAwindows;
	    d.text.TTinit();
	    d.text.watch = p.watch;
	    d.maptab = MTcgacolor;
	    d.setdispsize(d.text.nrows, d.text.ncols);
	    d.text.clear();
	    d.text.block_cursor();
	}
	if (plynum)
	    p.citsel();		// select city for each player
    }

    plynum = 1;			// get the default player
}

void winRestore()
{
    //PRINTF("winRestore()\n");
    debug
    {
	setran();		// seed random number generator
    }

    Text *t = &player[0].display.text;

    t.TTinit();

    for (plynum = 0; plynum <= numply; plynum++)
    {   Player *p = &player[plynum];
	Display *d = new Display();
	p.display = d;
	d.initialize();

	if (p.human)
	{
	    d.timeinterval = 1;
	}
	else
	{
	}

	if (plynum == 1)
	{
	    p.secflg = 1;
	    p.watch = DAwindows;
	    d.text.TTinit();
	    d.text.watch = p.watch;
	    d.maptab = MTcgacolor;
	    d.setdispsize(d.text.nrows, d.text.ncols);
	    d.text.clear();
	    d.text.block_cursor();
	}
    }
    plynum = 1;			// get the default player
}

/***************************************
 * Given new scale factors x,y, adjust sector location.
 * Returns:
 *	!=0 if sector moved
 */

int adjSector(double newscalex, double newscaley)
{
    int cursorx;
    int dx;
    int ncols;
    int scmin;
    int scmax;

    int cursory;
    int dy;
    int nrows;
    int srmin;
    int srmax;

    int gap;
    int n;
    int width;
    Display *d;

    int moved = 0;
    int newval;

    if (!global.player)		// if not initialized yet
	return 0;

//PRINTF("Before: ulcorner = %d\n", global.ulcorner);
    ncols = COL(global.cursor) - COL(global.ulcorner);
//PRINTF("cursor = %d, diff = %d, ncols = %d\n", global.cursor, global.cursor - global.ulcorner, ncols);
    dx = cast(int)(10 * global.scalex);
    cursorx = ncols * dx + (dx / 2) - global.offsetx;
    dx = cast(int)(10 * newscalex);
    n = (cursorx - (dx / 2) + (dx - 1)) / dx;
//PRINTF("n = %d, cursorx = %d, dx = %d\n", n, cursorx, dx);
    if (cursorx < dx + (dx / 2))
    {   n = 1;
	cursorx = n * dx + (dx / 2);
    }
    if (COL(global.cursor) < n)
    {   n = COL(global.cursor);
	cursorx = n * dx + (dx / 2);
    }
    width = (global.pixelx + (dx - 1)) / dx;
//PRINTF("n = %d, width = %d\n", n, width);

    // Right justify right edge
    gap = cursorx - dx/2 + 3 * dx - global.pixelx;
    if (gap > 0)
    {
	cursorx -= gap;
	n = (cursorx - (dx / 2) + (dx - 1)) / dx;
//PRINTF("1adj gap = %d\n", gap);
    }
    gap = global.pixelx - (((Mcolmx + 1) - COL(global.cursor)) * dx + cursorx - dx/2);
    if (gap > 0)
    {
	cursorx += gap;
	n = (cursorx - (dx / 2) + (dx - 1)) / dx;
//PRINTF("2adj gap = %d\n", gap);
	if (COL(global.cursor) < n)
	{   n = COL(global.cursor);
	    cursorx = n * dx + (dx / 2);
	}
    }

    newval = dx * n + dx / 2 - cursorx;
    if (newval != global.offsetx)
	moved = 1;
    global.offsetx = newval;
//PRINTF("offsetx = %d\n", global.offsetx);
    scmin = COL(global.cursor) - n;
    scmax = (width - 1);
//PRINTF("n = %d, width = %d, scmin = %d, COL(cursor) = %d\n", n, width, scmin, COL(global.cursor));
    assert(scmin <= COL(global.cursor));

    nrows = ROW(global.cursor) - ROW(global.ulcorner);
//PRINTF("cursor = %d, diff = %d, nrows = %d\n", global.cursor, global.cursor - global.ulcorner, nrows);
    dy = cast(int)(10 * global.scaley);
    cursory = nrows * dy + (dy / 2) - global.offsety;
    dy = cast(int)(10 * newscaley);
    n = (cursory - (dy / 2) + (dy - 1)) / dy;
//PRINTF("n = %d, cursory = %d, dy = %d\n", n, cursory, dy);
    if (cursory < dy + (dy / 2))
    {   n = 1;
	cursory = n * dy + (dy / 2);
    }
    if (ROW(global.cursor) < n)
    {   n = ROW(global.cursor);
	cursory = n * dy + dy / 2;
//PRINTF("adjust top\n");
    }
    width = (global.pixely + (dy - 1)) / dy;

    // Bottom justify bottom edge
    gap = cursory - dy/2 + 3 * dy - global.pixely;
    if (gap > 0)
    {
	cursory -= gap;
	n = (cursory - (dy / 2) + (dy - 1)) / dy;
    }
    gap = global.pixely - (((Mrowmx + 1) - ROW(global.cursor)) * dy + cursory - dy/2);
    if (gap > 0)
    {
	cursory += gap;
	n = (cursory - (dy / 2) + (dy - 1)) / dy;
	if (ROW(global.cursor) < n)
	{   n = ROW(global.cursor);
	    cursory = n * dy + dy / 2;
	}
    }

    newval = dy * n + dy / 2 - cursory;
    if (newval != global.offsety)
	moved = 1;
    global.offsety = newval;
    srmin = ROW(global.cursor) - n;
    srmax = (width - 1);
//PRINTF("n = %d, height = %d, srmin = %d\n", n, width, srmin);

    d = global.player.display;
    //d.Smin = srmin * 256 + scmin;
    d.Smax = srmax * 256 + scmax;

    newval = srmin * (Mcolmx + 1) + scmin;
    if (newval != global.ulcorner)
	moved = 1;
    global.ulcorner = newval;
    d.secbas = global.ulcorner;
//PRINTF("offsetx = %d, offsety = %d\n", global.offsetx, global.offsety);
//PRINTF("After: ulcorner = %d, Smin = %x, Smax = %x\n", global.ulcorner, d.Smin, d.Smax);

    assert(global.ulcorner >= 0 && global.ulcorner < MAPSIZE);
    assert(global.ulcorner <= global.cursor && global.cursor < MAPSIZE);
    assert(COL(global.ulcorner) <= COL(global.cursor));
    assert(dx > 0 && dy > 0);
    assert(global.offsetx >= 0 && global.offsetx < dx);
    assert(global.offsety >= 0 && global.offsety < dy);

    global.scalex = newscalex;
    global.scaley = newscaley;

    return moved;
}

/*************************************
 * Invalidate display of loc on the screen, so
 * it will get updated.
 */

void invalidateLoc(loc_t loc)
{

    Rect rect;
    int r, c;
    int dx;
    int dy;

//PRINTF("invalidateLoc(loc = %d)\n", loc);
    assert(loc < MAPSIZE);

    r = ROW(loc) - ROW(global.ulcorner);
    c = COL(loc) - COL(global.ulcorner);
    dx = cast(int)(10 * global.scalex);
    dy = cast(int)(10 * global.scaley);

    rect.LTWH(c * dx - global.offsetx,
	      4*global.cyChar+global.cyOffset + r * dy - global.offsety,
	      dx, dy);
    global.win.repaint(rect);
}

/*************************************
 * Invalidate display of rectangle formed by corners loc1, loc2
 * on the screen, so
 * it will get updated.
 */

void invalidateLocRect(loc_t loc1, loc_t loc2)
{
    Rect rect;
    int r1, c1;
    int r2, c2;
    int dx;
    int dy;

//PRINTF("invalidateLoc(loc = %d)\n", loc);
    assert(loc1 < MAPSIZE);
    assert(loc2 < MAPSIZE);

    r1 = ROW(loc1) - ROW(global.ulcorner);
    c1 = COL(loc1) - COL(global.ulcorner);
    r2 = ROW(loc2) - ROW(global.ulcorner);
    c2 = COL(loc2) - COL(global.ulcorner);

    if (r1 > r2)
    {	int r;
	r = r1;
	r1 = r2;
	r2 = r;
    }
    if (c1 > c2)
    {	int c;
	c = c1;
	c1 = c2;
	c2 = c;
    }

    dx = cast(int)(10 * global.scalex);
    dy = cast(int)(10 * global.scaley);

    rect.LTWH(c1 * dx - global.offsetx,
	      4*global.cyChar+global.cyOffset + r1 * dy - global.offsety,
	      dx * (c2 - c1 + 1),
	      dy * (r2 - r1 + 1));
    global.win.repaint(rect);
}

/******************************************
 * Invalidate entire sector.
 */

void invalidateSector()
{
  global.win.repaint(global.sector);
}

/******************************************
 * Convert loc to screen coordinate.
 */

int LocToX(loc_t loc)
{
    int dx;
    int x;
    int col;

    col = COL(loc) - COL(global.ulcorner);
    dx = cast(int)(10 * global.scalex);
    x = col * dx + dx / 2 - global.offsetx;
    return x;
}

int LocToY(loc_t loc)
{
    int dy;
    int y;
    int row;

    row = ROW(loc) - ROW(global.ulcorner);
    dy = cast(int)(10 * global.scaley);
    y = 4*global.cyChar+global.cyOffset + row * dy + dy / 2 - global.offsety;
    return y;
}

/*********************************************
 * Start/Stop blast graphic.
 */

void ShowBlast(int state, loc_t loc)
{
    Rect blastbox;
    int x, y;

    x = LocToX(loc);
    y = LocToY(loc);
    blastbox.LTRB(x - 10, blastbox.bottom - 20, x + 10,y + 5);
    global.win.repaint(blastbox);
    global.blastState = state;
    global.blastx = blastbox.left;
    global.blasty = blastbox.top;
    if (state)
      global.win.repaintNow();
}

class NewGameDialog : Dialog {
  this(Window parent) {
    super(parent,"New Game");
    layoutMgr = new FlowLayout;

    // until the bugs in GroupBox get worked out avoid using them
    /*
    GroupBox gb = new GroupBox(this,"Num Enemies");
    static double[2] yy = [.5,.5];
    static double[3] xx = [.33,.33,.33];
    gb.layoutMgr = new TableLayout(xx,yy);
    */
    Component gb = this;
    toggle = new ToggleGroup();
    RadioButton n1 = new RadioButton(gb,"0");
    toggle.addButton(n1);
    RadioButton n2 = new RadioButton(gb,"1");
    toggle.addButton(n2);
    RadioButton n3 = new RadioButton(gb,"2");
    toggle.addButton(n3);
    RadioButton n4 = new RadioButton(gb,"3");
    toggle.addButton(n4);
    RadioButton n5 = new RadioButton(gb,"4");
    toggle.addButton(n5);
    RadioButton n6 = new RadioButton(gb,"5");
    toggle.addButton(n6);
    toggle.select(2);

    demoButton = new CheckBox(this,"Demo");
    Button okButton = new Button(this,"OK");
    okButton.actionDelegate ~= &okCallback;
    Button cancelButton = new Button(this,"Cancel");
    cancelButton.actionDelegate ~= &cancelCallback;
    pack();
  }
  void okCallback(Component c) {
    ok = true;
    visible = false;
  }
  void cancelCallback(Component c) {
    ok = false;
    visible = false;
  }
  bool ok;
  CheckBox demoButton;
  ToggleGroup toggle;
  void getData(inout int demo, inout int num_players) {
    demo = demoButton.selected;
    num_players = IDD_ONE+toggle.selected; // use widget definition of num_players
  }
  void setData(inout int demo, inout int num_players) {
    demoButton.selected = demo != 0;
    toggle.select(num_players-IDD_ONE);
    toggle.buttons[num_players-IDD_ONE].requestFocus();
  }
}

class SelectCityDialog : Dialog {
  this(Window parent) {
    super(parent,"City Production");
    layoutMgr = new FlowLayout;

    // until the bugs in GroupBox get worked out avoid using them
    /*
    GroupBox gb = new GroupBox(this,"Num Enemies");
    static double[2] yy = [.5,.5];
    static double[3] xx = [.33,.33,.33];
    gb.layoutMgr = new TableLayout(xx,yy);
    */
    Component gb = this;
    toggle = new ToggleGroup();
    RadioButton n1 = new RadioButton(gb,"Armies");
    toggle.addButton(n1);
    RadioButton n2 = new RadioButton(gb,"Fighters");
    toggle.addButton(n2);
    RadioButton n3 = new RadioButton(gb,"Destroyers");
    toggle.addButton(n3);
    RadioButton n4 = new RadioButton(gb,"Transports");
    toggle.addButton(n4);
    RadioButton n5 = new RadioButton(gb,"Submarines");
    toggle.addButton(n5);
    RadioButton n6 = new RadioButton(gb,"Cruisers");
    toggle.addButton(n6);
    RadioButton n7 = new RadioButton(gb,"Carriers");
    toggle.addButton(n7);
    RadioButton n8 = new RadioButton(gb,"Battleships");
    toggle.addButton(n8);
    toggle.select(0);

    //TODO: pictures on the right

    Button okButton = new Button(this,"OK");
    okButton.actionDelegate ~= &okCallback;
    Button cancelButton = new Button(this,"Cancel");
    cancelButton.actionDelegate ~= &cancelCallback;
    pack();
  }
  void okCallback(Component c) {
    ok = true;
    visible = false;
  }
  void cancelCallback(Component c) {
    ok = false;
    visible = false;
  }
  bool ok;
  ToggleGroup toggle;
  void getData(inout int phase) {
    phase = IDD_ARMIES+toggle.selected; // use widget definition of phase
  }
  void setData(inout int phase) {
    toggle.select(phase-IDD_ARMIES);
    toggle.buttons[phase-IDD_ARMIES].requestFocus();
  }
}

void showAboutBox() {
  char[] str = "Empire 2.03" \n
    "Wargame of the Century" \n
    "(c) Walter Bright 1978-2004" \n
    "All Rights Reserved" \n
    "www.classicempire.com" \n
    "powered by Digital Mars D" \n
    "www.digitalmars.com/d/";
  informationDialog(global.win,str,"About");
}
